/* 
 * $Id: zfcp_zh.c,v 1.3.2.2 2004/06/08 10:10:29 mpeschke Exp $
 *
 * Module providing an interface for HBA API (FC-HBA) implementation
 * to the zfcp driver.
 *
 * (C) Copyright IBM Corp. 2002, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 *
 * Authors:
 *       Stefan Voelkel <Stefan.Voelkel@millenux.com>
 *       Andreas Herrmann <aherrman@de.ibm.com>
 */
#include "zfcp.h"
#include "zfcp_zh.h"

#include <linux/module.h>

#define ZFCP_LOG_AREA			ZFCP_LOG_AREA_OTHER
#define ZFCP_LOG_AREA_PREFIX		ZFCP_LOG_AREA_PREFIX_OTHER

extern zfcp_data_t zfcp_data;

struct zfcp_callbacks zfcp_callback = { };

/**
 * zfcp_callback_do_adapter_add - callback wrapper
 * @*filp: passed through to the callback
 * @*a: adapter that was added
 *
 * We do not need irq safe spinlocks here, since we onyl do a read_lock. All
 * writers must use the _irq version though.
 */
void zfcp_callback_do_adapter_add(struct file *filp, const zfcp_adapter_t *a)
{
	unsigned long flags;
	
	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->adapter_add != NULL)) {
		zfcp_callback.callbacks->adapter_add(filp, a->devno, a->wwnn,
						     a->wwpn);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_callback_do_port_add - callback wrapper
 * @*filp: passed through to the callback
 * @*a: adapter the port was added to
 * @*p: port that was added
 */
void zfcp_callback_do_port_add(struct file *filp, const zfcp_adapter_t *a,
			       const zfcp_port_t *p)
{
	unsigned long flags;

	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->port_add != NULL)) {
		zfcp_callback.callbacks->port_add(filp, a->devno, p->wwpn,
						  p->wwnn, p->d_id);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_callback_do_unit_add - callback wrapper
 * @*filp: passed through to the callback
 * @*a: Adapter the port belongs to
 * @*p: Port the unit belongs to
 * @*u: unit that was added
 */
void zfcp_callback_do_unit_add(struct file *filp, const zfcp_adapter_t *a,
			       const zfcp_port_t *p, const zfcp_unit_t *u)
{
	unsigned long flags;

	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->unit_add != NULL)) {
		zfcp_callback.callbacks->unit_add(filp, a->devno, p->wwpn,
				u->fcp_lun, a->scsi_host->host_no,
				u->device->channel, u->device->id,
				u->device->lun);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_callback_do_incomming_els - callback wrapper
 * @*a: adapter the ELS was recieved
 * @*v: pointer to the ELS payload
 */
void zfcp_callback_do_incomming_els(const zfcp_adapter_t *a, const void *v)
{
	unsigned long flags;

	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->incomming_els != NULL)) {
		zfcp_callback.callbacks->incomming_els(a->devno, a->s_id, v);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_callback_do_link_down - callback wrapper
 * @*a: adapter that lost the link
 */
void zfcp_callback_do_link_down(const zfcp_adapter_t *a)
{
	unsigned long flags;

	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->link_down != NULL)) {
		zfcp_callback.callbacks->link_down(a->s_id);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_callback_do_link_up - callback wrapper
 * @*a: adapter with link up event
 */
void zfcp_callback_do_link_up(const zfcp_adapter_t *a)
{
	unsigned long flags;

	read_lock_irqsave(&zfcp_callback.lock, flags);

	if ((zfcp_callback.callbacks != NULL) &&
	    (zfcp_callback.callbacks->link_up != NULL)) {
		zfcp_callback.callbacks->link_up(a->s_id);
	}

	read_unlock_irqrestore(&zfcp_callback.lock, flags);
}

/**
 * zfcp_search_unit - search for a unit
 * @*port: pointer to port structure of port where unit is attached to
 * @lun: FC LUN of the unit to search for
 * @**unit: address to write pointer to found unit structure to
 * Return: 0 on success, -E* code else
 * Locks: lock/unlock of port->unit_list_lock
 *
 * Search for an unit and return its address.
 * See also zfcp_search_port_and_unit().
 */
static int zfcp_search_unit(zfcp_port_t *port, fcp_lun_t lun,
			    zfcp_unit_t **unit)
{
	zfcp_unit_t *u;
	unsigned long flags;
	int ret = 0;

	*unit = NULL;

	read_lock_irqsave(&port->unit_list_lock, flags);
	ZFCP_FOR_EACH_UNIT(port, u)
	{
		if (u->fcp_lun == lun) {
			*unit = u;
		}
	}
	read_unlock_irqrestore(&port->unit_list_lock, flags);
	if (NULL == *unit) {
		ret = -ENOUNI;
	}
	
	return ret;
}

/**
 * zfcp_search_port_unit - search for a port and unit
 * @*adapter: pointer to adapter structure of adapter where port is attached to
 * @wwpn: of the port to search for
 * @lun: of the unit to search for, ignored if @**unit == NULL
 * @**port: address to write pointer to the found port structure to
 * @**unit: address to write pointer to the found unit structure to
 *	(iff @**unit != NULL)
 * Return: 0 on success, -E* code else
 * Locks: lock/unlock of adapter->port_list_lock
 *
 * Search for port and unit and return their addresses.
 * If @**unit == NULL search only for port.
 * If @**unit != NULL search also for unit.
 * See also zfcp_search_adapter_port_unit() and zfcp_search_unit().
 */
static int zfcp_search_port_unit(zfcp_adapter_t *adapter, wwn_t wwpn,
				 fcp_lun_t lun,  zfcp_port_t **port,
				 zfcp_unit_t **unit)
{
	zfcp_port_t *p;
	unsigned long flags;
	int ret = 0;

	*port = NULL;

	read_lock_irqsave(&adapter->port_list_lock, flags);
	ZFCP_FOR_EACH_PORT(adapter, p)
	{
		if (p->wwpn == wwpn) {
			*port = p;
		}
	}
	read_unlock_irqrestore(&adapter->port_list_lock, flags);

	if (NULL == *port) {
		ret = -ENOPOR;
	} else {
		if (NULL != unit) {
			ret = zfcp_search_unit(*port, lun, unit);
		}
	}

	return ret;
}

/**
 * zfcp_search_adapter_port_unit - search for an adapter, port and unit
 * @devno: of the adapter to search for
 * @wwpn: of the port to search for, ignored if **port == NULL
 * @lun: of the unit to search for, ignored if **port == NULL || **unit == NULL
 * @**adapter: address to write pointer to found adapter structure to
 * @**port: address to write pointer to found port structure to,
 *	(iff **port != NULL)
 * @**unit: address to write pointer to found unit structure to,
 *	(iff **port != NULL && **unit != NULL)
 * Return: 0 on success, -E* code else
 * Locks: lock/unlock of zfcp_data.adapter_list_lock
 *
 * Search for an adapter, port and unit and return their addresses.
 * If @**port == NULL, search only for an adapter.
 * If @**port != NULL, search also for port.
 * If @**port != NULL and @**unit != NULL search also for port and unit.
 * See also zfcp_search_port_unit().
 */
static int zfcp_search_adapter_port_unit(devno_t devno, wwn_t wwpn,
					 fcp_lun_t lun,
					 zfcp_adapter_t **adapter,
					 zfcp_port_t **port, zfcp_unit_t **unit)
{
	zfcp_adapter_t *a;
	unsigned long flags;
	int ret = 0;

	if (NULL == adapter) {
		return -EINVAL;
	}

	*adapter = NULL;
	
	read_lock_irqsave(&zfcp_data.adapter_list_lock, flags);
	ZFCP_FOR_EACH_ADAPTER(a)
	{
		if (a->devno == devno) {
			if (!atomic_test_mask(ZFCP_STATUS_COMMON_ERP_FAILED,
					      &a->status) &&
			    atomic_test_mask((ZFCP_STATUS_COMMON_RUNNING |
					      ZFCP_STATUS_COMMON_UNBLOCKED),
					     &a->status)) {
				*adapter = a;
			}
			break;
		}
	}
	read_unlock_irqrestore(&zfcp_data.adapter_list_lock, flags);
	if (NULL == *adapter) {
		ret = -ENOADA;
	} else {
		if (NULL != port) {
			ret = zfcp_search_port_unit(*adapter, wwpn, lun,
						    port, unit);
		}
	}

	return ret;
}

/**
 * get_config_units - create unit config events
 * @*filp: passed through to callback
 * @*adapter: the unit belongs to
 * @*port: the unit belongs to
 *
 * generate one unit add event for each unit below the passed port
 */
static inline int get_config_units(struct file *filp, zfcp_adapter_t *adapter,
				   zfcp_port_t *port)
{
	zfcp_unit_t *unit;
	unsigned long flags;
	int ret = 0;
	
	read_lock_irqsave(&port->unit_list_lock, flags);

	ZFCP_FOR_EACH_UNIT(port, unit)
	{
		if (unit->device != NULL) {
			zfcp_callback_do_unit_add(filp, adapter, port, unit);
			++ret;
		}
	}

	read_unlock_irqrestore(&port->unit_list_lock, flags);

	return ret;
}

/**
 * get_config_ports - create port config events, or search for a port.
 * @*filp: passed through to callback
 * @*adapter: the port belongs to
 * @wwpn: generate port add events if 0, or search port and get_config_units()
 */
static inline int get_config_ports(struct file *filp, zfcp_adapter_t *adapter,
				   wwn_t wwpn, unsigned int config_flags)
{
	zfcp_port_t *port;
	unsigned long flags;
	int ret = 0;

	read_lock_irqsave(&adapter->port_list_lock, flags);

	ZFCP_FOR_EACH_PORT(adapter, port)
	{
		if (ZH_GET_CONFIG_PORTS == config_flags) {
			/* ignore name server port */
			if (0 != port->wwpn) {
				zfcp_callback_do_port_add(filp, adapter, port);
				++ret;
			}
		} else {
			if (port->wwpn != wwpn) {
				ret = -ENOPOR;
			} else {
				ret = get_config_units(filp, adapter, port);
				break;
			}
		}
	}

	read_unlock_irqrestore(&adapter->port_list_lock, flags);

	return ret;
}

/**
 * zfcp_zh_get_config - Prepare config for userspace
 * @*filp: passed through to callback
 * @devno: generate adapter add events if 0, or search and get_config_ports()
 * @wwpn: passed to get_config_ports()
 *
 * TNG get_config method, uses the callback methods and so generates
 * events for each configured item. The events are private to the
 * passed file descriptor.
 */
int zfcp_zh_get_config(struct file *filp, devno_t devno, wwn_t wwpn,
		unsigned int config_flags)
{
	zfcp_adapter_t *adapter;
	unsigned long flags;
	int ret = 0;

	read_lock_irqsave(&zfcp_data.adapter_list_lock, flags);

	ZFCP_FOR_EACH_ADAPTER(adapter)
	{
		if (ZH_GET_CONFIG_ADAPTERS == config_flags) {
			zfcp_callback_do_adapter_add(filp, adapter);
			++ret;
		} else {
			if (adapter->devno != devno) {
				ret = -ENOADA;
			} else {
				ret = get_config_ports(filp, adapter, wwpn,
						       config_flags);
				break;
			}
				
		}
	}
		
	read_unlock_irqrestore(&zfcp_data.adapter_list_lock, flags);

	return ret;
}

/**
 * zfcp_zh_get_adapter_attributes - provide data for the api call
 * @devno: of the adapter 
 * @attr: pointer to struct zfcp_adapter_attributes to return attributes
 * Return: 0 on success, -E* else
 * Context: user
 * Locks: lock/unlock zfcp_data.adapter_list_lock
 */
int zfcp_zh_get_adapter_attributes(devno_t devno,
				   struct zfcp_adapter_attributes *attr)
{
	zfcp_adapter_t *adapter;
	int ret = 0;
	
	memset(attr, 0, sizeof(*attr));

	ret = zfcp_search_adapter_port_unit(devno, 0, 0, &adapter, NULL, NULL);
	if (ret != 0) {
		return ret;
	}

	strcpy(attr->manufacturer, "IBM");
	strncpy(attr->serial_number, adapter->serial_number, 32);
	switch (adapter->hydra_version) {
	case FSF_ADAPTER_TYPE_FICON:
		strcpy(attr->model, "FICON FCP");
		break;
	case FSF_ADAPTER_TYPE_FICON_EXPRESS:
		strcpy(attr->model, "FICON Express FCP");
		break;
	}
	strcpy(attr->model_description, "zSeries Fibre Channel Adapter");
	attr->node_wwn = adapter->wwnn;
	sprintf(attr->hardware_version, "0x%x",
		adapter->hardware_version);
	sprintf(attr->driver_version, "0x%x", zfcp_data.driver_version);
	sprintf(attr->firmware_version, "0x%x",
		adapter->fsf_lic_version);
	attr->vendor_specific_id = 42;
	attr->number_of_ports = 1;
	strcpy(attr->driver_name, "zfcp.o");
	/* option_rom_version not used, node_symbolic_name not set */

	return ret;
}

/**
 * zfcp_zh_get_port_statistics - Retrieve statistics of adapter port
 * @devno: adapter of which port data should be reported
 * @stat: pointer to struct zfcp_port_statistics to return statistics
 * Return: 0 on success,  -E* else
 * Context: user
 * Locks: lock/unlock zfcp_data.adapter_list_lock
 */
int zfcp_zh_get_port_statistics(devno_t devno,
				struct zfcp_port_statistics *stat)
{
	zfcp_adapter_t *adapter;
	fsf_qtcb_bottom_port_t data;
	int ret = 0;

	memset(stat, 0, sizeof(*stat));
	
	ret = zfcp_search_adapter_port_unit(devno, 0, 0, &adapter, NULL, NULL);
	if (ret != 0) {
		return ret;
	}

	if(adapter->supported_features & FSF_FEATURE_HBAAPI_MANAGEMENT){
		ret = zfcp_fsf_exchange_port_data(adapter, &data);
		if (0 == ret) {
			/* convert fsf_qtcb_bottom_port into
			   zfcp_port_statistics */
			stat->last_reset = data.seconds_since_last_reset;
			stat->tx_frames = data.tx_frames;
			stat->tx_words = data.tx_words;
			stat->rx_frames = data.rx_words;
			stat->lip = data.lip;
			stat->nos = data.nos;
			stat->error_frames = data.error_frames;
			stat->dumped_frames = data.dumped_frames;
			stat->link_failure = data.link_failure;
			stat->loss_of_sync = data.loss_of_sync;
			stat->loss_of_signal = data.loss_of_signal;
			stat->prim_seq_prot_error = data.psp_error_counts;
			stat->invalid_tx_words = data.invalid_tx_words;
			stat->invalid_crc = data.invalid_crcs;
			stat->input_requests = data.input_requests;
			stat->output_requests = data.output_requests;
			stat->control_requests = data.control_requests;
			stat->input_megabytes = data.input_mb;
			stat->output_megabytes = data.output_mb;
		}
	} else {
		ret = -EOPNOTSUPP;
	}

	return ret;
}


/**
 * zfcp_zh_get_port_attributes - Retrieve attributes of adapter port
 * @devno: adapter of which port data should be reported
 * @attr: pointer to struct zfcp_port_attributes to return attributes
 * Return: 0 on success,  -E* else
 * Context: user
 * Locks: lock/unlock zfcp_data.adapter_list_lock, adapter->port_list_lock
 */
int zfcp_zh_get_port_attributes(devno_t devno,
				struct zfcp_port_attributes *attr)
{
	zfcp_adapter_t *adapter;
	zfcp_port_t *port;
	fsf_qtcb_bottom_port_t data;
	unsigned long flags;
	int ret = 0;

	memset(attr, 0, sizeof(*attr));
	
	ret = zfcp_search_adapter_port_unit(devno, 0, 0, &adapter, NULL, NULL);
	if (ret != 0) {
		return ret;
	}

	attr->wwnn = adapter->wwnn;
	attr->wwpn = adapter->wwpn;
	attr->speed = adapter->fc_link_speed;
	attr->discovered_ports = adapter->ports;

	/* ignore nameserver port */
	read_lock_irqsave(&adapter->port_list_lock, flags);
	ZFCP_FOR_EACH_PORT(adapter, port)
	{
		if(port->wwpn == 0){
			if(attr->discovered_ports)
				--attr->discovered_ports;
			break;
		}
	}
	read_unlock_irqrestore(&adapter->port_list_lock, flags);

	if(adapter->supported_features & FSF_FEATURE_HBAAPI_MANAGEMENT){
		ret = zfcp_fsf_exchange_port_data(adapter, &data);
		if (0 == ret) {
			/* convert fsf_qtcb_bottom_port into
			   zfcp_port_attributes */
			attr->fcid = data.fc_port_id;
			attr->type = data.port_type;
			attr->state = data.port_state;
			attr->supported_class_of_service =
				data.class_of_service;
			attr->supported_speed = data.supported_speed;
			attr->max_frame_size = data.maximum_frame_size;
			memcpy(&attr->supported_fc4_types,
			       &data.supported_fc4_types, 32);
			memcpy(&attr->active_fc4_types,
			       &data.active_fc4_types, 32);
		}
	} else {
		ret = -EOPNOTSUPP;
	}
	/* fabric_name and symbolic_name not set */

	return ret;
}

/**
 * zfcp_zh_get_discovered_port_attributes - Retrieve attributes of target port
 * @devno: adapter for which port data should be reported
 * @wwpn: wwn of discovered port
 * @attr: pointer to struct zfcp_port_attributes to return attributes
 * Return: 0 on success,  -E* else
 * Context: user
 * Locks: lock/unlock zfcp_data.adapter_list_lock, adapter->port_list_lock
 */
int zfcp_zh_get_dport_attributes(devno_t devno, wwn_t wwpn,
				 struct zfcp_port_attributes *attr)
{
	zfcp_adapter_t *adapter;
	zfcp_port_t *port;
	struct ct_iu_ga_nxt *ct_iu_resp;
	int ret;

	memset(attr, 0, sizeof(*attr));

	ret = zfcp_search_adapter_port_unit(devno, wwpn, 0,
					    &adapter, &port, NULL);
	if (ret != 0) {
		return ret;
	}

	ct_iu_resp = kmalloc(sizeof(*ct_iu_resp), GFP_KERNEL);
	if (0 == ct_iu_resp) {
		return -ENOMEM;
	}

	ret = zfcp_ns_ga_nxt_request(port, ct_iu_resp);
	if (0 == ret) {
		attr->wwnn = port->wwnn;
		attr->wwpn = port->wwpn;
		attr->fabric_name = (wwn_t) ct_iu_resp->du.fabric_wwn;
		attr->fcid = port->d_id;

		/* map FC-GS-2 port types to HBA API
		   port types */
		switch(ct_iu_resp->du.port_type) {
		case 0x01: /* N_Port */
			attr->type = FSF_HBA_PORTTYPE_NPORT;
		case 0x02: /* NL_Port */
			attr->type = FSF_HBA_PORTTYPE_NLPORT;
		case 0x81: /* F_Port */
			attr->type = FSF_HBA_PORTTYPE_FPORT;
		case 0x82: /* FL_Port */
			attr->type = FSF_HBA_PORTTYPE_FLPORT;
		case 0x03: /* F/NL_Port */
		case 0x7f: /* Nx_Port */
		case 0x84: /* E_Port */
			attr->type = FSF_HBA_PORTTYPE_OTHER;
		case 0x00: /* Unidentified */
		default: /* reserved */
			attr->type = FSF_HBA_PORTTYPE_UNKNOWN;
		}

		attr->state = FSF_HBA_PORTSTATE_UNKNOWN;
		attr->supported_class_of_service = ct_iu_resp->du.cos;
		memcpy(&attr->active_fc4_types, &ct_iu_resp->du.fc4_types, 32);
		memcpy(&attr->symbolic_name,
		       &ct_iu_resp->du.node_symbolic_name,
		       ct_iu_resp->du.port_symbolic_name_length);
	}
	kfree(ct_iu_resp);

	/* supported_speed, speed, max_frame_size, supported_fc4_types,
	   discovered_ports not set */

	return ret;
}


/**
 * zfcp_zh_send_ct_handler() - handler for zfcp_zh_send_ct()
 * @data: a pointer to struct zfcp_send_ct, It was set as handler_data
 *	in zfcp_zh_send_ct().
 * Context: interrupt
 *
 * This handler is called on completion of a send_ct request. We just wake up
 * our own zfcp_zh_send_ct() function.
 */
static void zfcp_zh_send_ct_handler(unsigned long data)
{

        struct zfcp_send_ct *ct = (struct zfcp_send_ct *) data;
	struct completion *wait = (struct completion*) ct->completion;

	complete(wait);
	return;
}


/**
 * zfcp_zh_send_ct() - send a CT_IU containing FC-GS-4 command
 * @devno: adapter for which port data should be reported
 * @req: scatter-gather list with request data
 * @req_count: number of elements in @req
 * @resp: scatter-gather list for response data
 * @resp_count: number of elements in @resp
 * Return: 0 on success,  -E* else
 * Context: user
 * Locks: lock/unlock zfcp_data.adapter_list_lock
 *
 * Note: Currently only requests to the nameserver port are supported.
 *	Other well-known ports currently have no representation in zfcp.
 */
int zfcp_zh_send_ct(devno_t devno,
		    struct scatterlist *req, unsigned int req_count,
		    struct scatterlist *resp, unsigned int resp_count)
{
	struct zfcp_send_ct ct;
	zfcp_adapter_t *adapter;
	struct ct_hdr *ct_header;
	int ret;

	DECLARE_COMPLETION(wait);

	memset(&ct, 0, sizeof(ct));

	ret = zfcp_search_adapter_port_unit(devno, 0, 0, &adapter, NULL, NULL);
	if (ret != 0) {
		return ret;
	}

	ct_header = (struct ct_hdr *) req[0].address;
	if ((ct_header->gs_type != ZFCP_CT_DIRECTORY_SERVICE) ||
	    (ct_header->gs_subtype != ZFCP_CT_NAME_SERVER)) {
		/* currently only nameserver requests are supported */
		ZFCP_LOG_NORMAL("Tried to send CT IU to other service than "
				"Name Server Directory Service. This is "
				"currently not supported.\n");
		return -ENOPOR;
	}

	if (!adapter->nameserver_port) {
		ZFCP_LOG_NORMAL("Name Server port not available\n.");
		return -ENOPOR;
	}
	ct.port = adapter->nameserver_port;

	ct.req = req;
	ct.resp = resp;
	ct.req_count = req_count;
	ct.resp_count = resp_count;
	ct.handler = zfcp_zh_send_ct_handler;
	ct.handler_data = (unsigned long) &ct;
	ct.completion = &wait;
	ct.timeout = ZFCP_CT_TIMEOUT;

	ret = zfcp_fsf_send_ct(&ct, 0, 0);
	if (0 == ret) {
		wait_for_completion(&wait);
	}

	return ret;
}

/**
 * zfcp_zh_callbacks_register - register callbacks of zfcp_hbaapi
 * @*cb: pointer to the callback structure
 * Return: 0 on success, else -E* code
 * Locks: lock/unlock of zfcp_callback.lock
 * Context: user
 */
int zfcp_zh_callbacks_register(struct zfcp_zh_callbacks *cb)
{
	unsigned long flags;
	int ret = 0;
	
	write_lock_irqsave(&zfcp_callback.lock, flags);

	if (zfcp_callback.callbacks != NULL) {
		ret = -EBUSY;
	} else {
		zfcp_callback.callbacks = cb;
	}

	write_unlock_irqrestore(&zfcp_callback.lock, flags);

	return ret;
}

/**
 * zfcp_zh_callbacks_unregister - unregister callbacks for zfcp_hbaapi
 * @*cb: pointer to the callback structure
 * Return: 0 on success, -E* code else
 * Locks: lock/unlock of zfcp_callback.lock
 * Context: user
 */
int zfcp_zh_callbacks_unregister(struct zfcp_zh_callbacks *cb)
{
	unsigned long flags;
	int ret = 0;

	write_lock_irqsave(&zfcp_callback.lock, flags);

	if (zfcp_callback.callbacks == NULL) {
		ret = -EBUSY;
	} else {
		if (zfcp_callback.callbacks != cb) {
			ZFCP_LOG_DEBUG("Tried to unregister callbacks from "
				       "different address.\n");
			ret = -EINVAL;
		} else {
			zfcp_callback.callbacks = NULL;
		}
	}

	write_unlock_irqrestore(&zfcp_callback.lock, flags);

	return ret;
}

/**
 * zfcp_zh_assert_fclun_zero - Assert that there is a FC LUN 0
 * @devno: devno of the adapter
 * @wwpn: wwpn of the discovered port
 *
 * Look for an unit at the passed adapter:port with FC LUN 0.
 * Add it if it does not exist. This unit is needed for
 * REPORT_LUNS.
 *
 * Note: No unit with a FC LUN 0 can be added for the same adpater and port
 *	after this call. (We could add an "overwriteable" flag to the
 *	zfcp_unit_t structure as a work-around for this.)
 */
int zfcp_zh_assert_fclun_zero(devno_t devno, wwn_t wwpn)
{
	int ret;
	zfcp_config_record_t cfg;
	zfcp_adapter_t *adapter;
	zfcp_port_t *port;
	zfcp_unit_t *unit;

	ret = zfcp_search_adapter_port_unit(devno, wwpn, 0,
					    &adapter, &port, &unit);
	if (ret != -ENOUNI) {
		return ret;
	}

	memset (&cfg, 0, sizeof(cfg));
	cfg.devno = devno;
	cfg.scsi_id = port->scsi_id;
	cfg.wwpn = port->wwpn;
	cfg.scsi_lun = port->max_scsi_lun +1;

	ret = zfcp_config_parse_record_add(&cfg);
	if (0 != ret) {
		return ret;
	}

	/* wait until the unit is opened */
	return zfcp_erp_wait(port->adapter);
}

/** 
 * zfcp_zh_send_scsi - Send a SCSI command to an unit
 * @devno: devno of the adapter
 * @wwpn: WWPN of the discovered port the unit is attached to
 * @lun: FC LUN of the unit to send the command to
 * @cmnd: address of the prepared Scsi_Cmnd
 * Return: 0 on success, > 0 on SCSI error, -E* code else
 *
 * FC LUN 0 is handled with extra care, to be able to send REPORT_LUNS.
 */
int zfcp_zh_send_scsi(devno_t devno, wwn_t wwpn, fcp_lun_t lun,
		      Scsi_Cmnd *cmnd)
{
	int ret;
	zfcp_adapter_t *adapter;
	zfcp_port_t *port;
	zfcp_unit_t *unit;
	
	ret = zfcp_search_adapter_port_unit(devno, wwpn, lun,
					    &adapter, &port, &unit);
	if (ret != 0) {
		return ret;
	}

	ret = zfcp_scsi_command_sync(unit, cmnd);
	if (ret < 0) {
		return ret;
	} else {
		return cmnd->result;
	}
}

/**
 * zfcp_zh_send_els - send an els to a port
 * @devno: of the adapter to send via
 * @wwpn: of the port to send to
 * @send: scatterlist describing the els payload to be sent
 * @send_count: number of elements in the send scatterlist
 * @receive: scatterlist describing buffers for the reply payload
 * @receive_count: number of elements in the receive scatterlist
 * Return: 0 on success, -E* code else
 * Locks: lock/unlock of zfcp_data.adapter_list_lock, adapter->port_list_lock
 */
int zfcp_zh_send_els(devno_t devno, wwn_t wwpn, struct scatterlist *send,
		     unsigned int send_count, struct scatterlist *receive,
		     unsigned int receive_count)
{
	int ret;
	struct zfcp_send_els *req;
	zfcp_adapter_t *adapter;
	zfcp_port_t *port;

	ret = zfcp_search_adapter_port_unit(devno, wwpn, 0,
					    &adapter, &port, NULL);
	if (ret != 0) {
		return ret;
	}

	req = kmalloc(sizeof(*req), GFP_KERNEL);
	if (NULL == req) {
		ret = -ENOMEM;
		goto out;
	}
	memset(req, 0, sizeof(*req));

	req->port = port;
	req->req = send;
	req->req_count = send_count;
	req->resp = receive;
	req->resp_count = receive_count;

	ret = zfcp_fsf_send_els_sync(req);

	if (0 == ret) {
		ret = req->status;
	}

	kfree(req);

	goto out;

out:
	return ret;
}

EXPORT_SYMBOL(zfcp_zh_callbacks_register);
EXPORT_SYMBOL(zfcp_zh_callbacks_unregister);
EXPORT_SYMBOL(zfcp_zh_get_config);
EXPORT_SYMBOL(zfcp_zh_get_adapter_attributes);
EXPORT_SYMBOL(zfcp_zh_get_port_attributes);
EXPORT_SYMBOL(zfcp_zh_get_port_statistics);
EXPORT_SYMBOL(zfcp_zh_get_dport_attributes);
EXPORT_SYMBOL(zfcp_zh_send_ct);
EXPORT_SYMBOL(zfcp_zh_send_els);
EXPORT_SYMBOL(zfcp_zh_send_scsi);
EXPORT_SYMBOL(zfcp_zh_assert_fclun_zero);
